import logging
import os
from pathlib import Path
from typing import MutableSequence, Union

from dictIO import CppDict, DictReader, DictWriter, create_target_file_name


__ALL__ = ['DictParser']

logger = logging.getLogger(__name__)


class DictParser():
    """Parser for dictionaries in dictIO dict file format, as well as JSON and XML

    DictParser is a convenience class.
    DictParser.parse() combines the operations of DictReader.read() and DictWriter.write() .
    """

    def __init__(self):
        return

    @staticmethod
    def parse(
        source_file: Union[str, os.PathLike[str]],
        includes: bool = True,
        mode: str = 'w',
        order: bool = False,
        comments: bool = True,
        scope: Union[MutableSequence[str], None] = None,
        output: Union[str, None] = None,
    ) -> Union[CppDict, None]:
        """Parses a dictionary file and saves it with prefix 'parsed.'

        DictParser.parse() combines the otherwise atomic operations
        of DictReader.read() and DictWriter.write() in one chunk:

        1: parsed_dict = DictReader.read(source_file)

        2: DictWriter.write(parsed_dict, target_file)

        The parsed dict is saved with prefix 'parsed.'
        Example: Parsing source file 'xyz' will result in parsed file 'parsed.xyz' being generated.

        The parsed dict will by default be written in dictIO's default dict file format.
        Optionally, output format can be changed to JSON, XML and OpenFOAM.

        Parameters
        ----------
        source_file : Union[str, os.PathLike[str]]
            dict file to be parsed
        includes : bool, optional
            merge sub-dicts being referenced through #include directives, by default True
        mode : str, optional
            append to output file ('a') or overwrite output file ('w'), by default 'w'
        order : bool, optional
            sort the parsed dict, by default False
        comments : bool, optional
            writes comments to output file, by default True
        scope : MutableSequence[str], optional
            scope the dict will be reduced to after parsing, by default None
        output : str, optional
            format of the output file. Choices are 'cpp', 'foam', 'xml' and 'json'., by default None

        Returns
        -------
        Union[CppDict, None]
            the parsed dict

        Raises
        ------
        FileNotFoundError
            if source_file does not exist
        """

        # Make sure source_file argument is of type Path. If not, cast it to Path type.
        source_file = source_file if isinstance(source_file, Path) else Path(source_file)
        if not source_file.exists():
            logger.error(f"DictParser: File {source_file} not found.")
            raise FileNotFoundError(source_file)

        logger.info(f"Parse {source_file}..")

        # Read (parse) the dict file
        parsed_dict = DictReader.read(
            source_file=source_file,
            includes=includes,
            order=order,
            comments=comments,
            scope=scope,
        )
        # Create filename for the parsed dict
        target_file = create_target_file_name(source_file, 'parsed', scope, output)
        # Save the parsed dict as a default dict file
        DictWriter.write(
            source_dict=parsed_dict,
            target_file=target_file,
            mode=mode,
            order=order,
        )

        logger.info(f"Successfully parsed {source_file} and saved as {target_file}.")

        return parsed_dict



import contextlib
import re
import os
from collections import UserDict
from pathlib import Path
from typing import (Any, Mapping, MutableMapping, MutableSequence, TypeVar, Union)
import logging
import dictIO
from dictIO.utils.counter import BorgCounter
from dictIO.utils.path import relative_path


__ALL__ = ['CppDict', 'order_keys', 'find_global_key', 'set_global_key', 'global_key_exists']

_KT = TypeVar('_KT')    # generic Type variable for keys
_VT = TypeVar('_VT')    # generic Type variable for values

logger = logging.getLogger(__name__)


class CppDict(UserDict):
    """Data structure for generic dictionaries

    CppDict inherits from UserDict. It can hence be used transparently also in a context
    where a dict or any other MutableMapping type is expected.
    """

    def __init__(self, file: Union[str, os.PathLike[str], None] = None):
        super().__init__()  # call base class constructor

        self.counter = BorgCounter()
        self.source_file = None
        self.path = Path.cwd()
        self.name = ''

        if file:
            # Make sure file argument is of type Path. If not, cast it to Path type.
            file = file if isinstance(file, Path) else Path(file)
            self.source_file = file.absolute()
            self.path = self.source_file.parent
            self.name = self.source_file.name

        self.line_content = []
        self.block_content = ''
        self.tokens = []
        self.string_literals = {}

        self.line_comments = {}
        self.block_comments: MutableMapping[int, str] = {}
        self.expressions = {}
        self.includes = {}

        self.brackets = [('{', '}'), ('[', ']'), ('(', ')'), ('<', '>')]
        self.delimiters = ['{', '}', '(', ')', '<', '>', ';', ',']
        self.openingBrackets = [
            '{', '[', '('
        ]                           # Note: < and > are not considered brackets, but operators used in filter expressions
        self.closingBrackets = ['}', ']', ')']
        return

    def include(self, dict_to_include: 'CppDict'):
        """Adds an include directive for the passed in dict

        Parameters
        ----------
        dict_to_include : CppDict
            The dict to be included via an include directive

        Raises
        ------
        AttributeError
            If dict_to_include.source_file is None
        ValueError
            If no relative path in between the current dict and the included dict can be resolved
        """
        if not dict_to_include.source_file:
            raise AttributeError(
                f"Cannot include {dict_to_include.name}. Attribute '.source_file' of {dict_to_include.name} is None."
            )
        if not self.source_file:
            raise AttributeError(
                f"Cannot include {dict_to_include.name}. Attribute '.source_file' of {self.name} is None."
            )

        include_file_path = dict_to_include.source_file
        relative_file_path: Path
        try:
            relative_file_path = relative_path(
                from_path=self.source_file.parent,
                to_path=dict_to_include.source_file,
            )
        except ValueError as e:
            raise ValueError(
                f"Cannot include {dict_to_include.name}. Relative path to {dict_to_include.name} could not be resolved."
            ) from e

        formatter = dictIO.formatter.CppFormatter()
        include_file_name = str(relative_file_path)
        include_file_name = include_file_name.replace('\\', '\\\\')
        include_file_name = formatter.format_type(include_file_name)

        include_directive = f'#include {include_file_name}'

        ii: int = 0
        placeholder: str = ''
        while True:
            ii = self.counter()
            placeholder = 'INCLUDE%06i' % ii
            if placeholder in self.data:
                continue
            else:
                break
        self.data[placeholder] = placeholder
        self.includes.update({ii: (include_directive, include_file_name, include_file_path)})
        return

    def update(self, __m: Mapping, **kwargs) -> None:
        """Updates top-level keys with the keys from the passed in dict.

        Overrides the update() method of UserDict base class in order to include also CppDict class attributes in the update.

        If a key already exists, it will be substituted by the key from the passed in dict.
        In order to not substitute top-level keys but recursively merge (potentially nested) content from passed in dict into the existing, use merge() instead.

        Note:

        The behaviour of update() corresponds with default mode '-w' in the dictParser command line interface.

        The behaviour of merge() corresponds with mode '-a' in the dictParser command line interface. See also CLI Documentation.

        Parameters
        ----------
        __m : Mapping
            dict containing the keys to be updated and its new values
        """
        # sourcery skip: class-extract-method
        # call base class method. This takes care of updating self.data
        super().update(__m, **kwargs)
        # update attributes
        if isinstance(__m, CppDict):
            self.line_comments.update(__m.line_comments)
            self.block_comments.update(__m.block_comments)
            self.expressions.update(__m.expressions)
            self.includes.update(__m.includes)
        self._clean()

        return

    def merge(self, dict: MutableMapping):
        """Merges the passed in dict into the existing CppDict instance.

        In contrast to update(), merge() works recursively. That is, it does not simply substitute top-level keys but
        recursively merges (potentially nested) content from the passed in dict into the existing.
        This prevents nested keys from being deleted.
        Further, existing keys will NOT be overwritten.

        Parameters
        ----------
        dict : MutableMapping
            dict to be merged
        """
        # merge dict into self (=into self.data)
        _merge_dicts(self, dict)
        # @TODO: An alternative we might test one day is the mergedeep module from the Python standard library:
        #        from mergedeep import merge
        #        self.data = merge(self.data, dict)
        #        CLAROS (FRALUM), 2022-01-05

        # merge CppDict attributes
        if isinstance(dict, CppDict):
            _merge_dicts(self.line_comments, dict.line_comments)
            _merge_dicts(self.block_comments, dict.block_comments)
            _merge_dicts(self.expressions, dict.expressions)
            _merge_dicts(self.includes, dict.includes)
        self._clean()

        return

    def __str__(self):
        """string representation of the CppDict instance in dictIO dict file format

        Returns
        -------
        str
            the string representation
        """
        from dictIO import \
            CppFormatter  # __str__ shall be formatted in default dict file format
        formatter = CppFormatter()
        return formatter.to_string(self)

    def __repr__(self):
        return f'CppDict({self.source_file!r})'

    def __eq__(self, other):
        return str(self) == str(other) if isinstance(other, CppDict) else False

    def order_keys(self):
        """alpha-numeric sorting of keys, recursively
        """
        self.data = dict(order_keys(self.data))
        self.line_comments = order_keys(self.line_comments)
        self.block_comments = order_keys(self.block_comments)
        self.expressions = order_keys(self.expressions)
        self.includes = order_keys(self.includes)

        return

    def find_global_key(self, query: str = '') -> Union[MutableSequence, None]:
        """Returns the global key thread to the first key the value of which matches the passed in query.

        Function works recursively on nested dicts and is non-greedy: The key of the first match is returned.
        Return value is a sequence of keys: The 'global key thread'.
        It represents the sequence of keys that one needs to traverse downwards
        in order to arrive at the target key found.

        Parameters
        ----------
        query : str, optional
            query string for the value to search for by default ''

        Returns
        -------
        Union[MutableSequence, None]
            global key thread to the first key the value of which matches the passed in query, if found. Otherwise None.
        """
        return find_global_key(self.data, query)

    def set_global_key(self, global_key: MutableSequence, value: Any = None):
        """Sets the value for the passed in global key.

        The global key thread is traversed downwards until arrival at the target key,
        the value of which is then set.

        Parameters
        ----------
        global_key : MutableSequence
            list of keys defining the global key thread to the target key (such as returned by method find_global_key())
        value : Any, optional
            value the target key shall be set to, by default None
        """
        set_global_key(self.data, global_key, value)

        return

    def global_key_exists(self, global_key: MutableSequence) -> bool:
        """Checks whether the specified global key exists.

        Parameters
        ----------
        global_key : MutableSequence
            global key the existence of which is checked

        Returns
        -------
        bool
            True if the specified global key exists, otherwise False
        """
        '''
        probe the existence of (nested) keys in dict
        '''
        return global_key_exists(self.data, global_key)

    def reduce_scope(self, scope: MutableSequence[str]):
        """Reduces the dict to the keys defined in scope.

        Parameters
        ----------
        scope : MutableSequence[str]
            scope the dict shall be reduced to
        """
        if scope:
            try:
                self.data = eval('self.data[\'' + '\'][\''.join(scope) + '\']')
            except KeyError as e:
                logger.warning(
                    'CppDict.reduce_scope(): no scope \'%s\' in dictionary %s' %
                    (e.args[0], self.source_file)
                )
        return

    @property
    def variables(self):
        variables = {}

        def extract_variables_from_dict(dict: MutableMapping):
            for k, v in dict.items():
                if isinstance(v, MutableMapping):
                    extract_variables_from_dict(v)      # recursion
                elif isinstance(v, MutableSequence):
                    if list_contains_dict(v):
                        extract_variables_from_list(v)  # recursion
                    else:
                                                        # special case: item is a list, but does NOT contain a nested dict (-> e.g. a vector or matrix)
                        variables.update({k: v})
                else:
                                                        # base case: item is a single value type
                    v = _insert_expression(v, self)
                    if not _value_contains_circular_reference(k, v):
                        variables.update({k: v})
            return

        def extract_variables_from_list(list: MutableSequence):
            # sourcery skip: remove-redundant-pass
            for v in list:
                if isinstance(v, MutableMapping):
                    extract_variables_from_dict(v)  # recursion
                elif isinstance(v, MutableSequence):
                    extract_variables_from_list(v)  # recursion
                else:
                                                    # By convention, list items are NOT added to the variables lookup table
                                                    # as they only have an index but no key
                                                    # (which we need, though, to serve as variable name)
                    pass
            return

        def list_contains_dict(list: MutableSequence) -> bool:
            # sourcery skip: merge-duplicate-blocks, use-any
            for item in list:
                if isinstance(item, MutableMapping):
                    return True
                elif isinstance(item, MutableSequence):
                    if list_contains_dict(item):
                        return True
            return False

        extract_variables_from_dict(self.data)

        order_keys(variables)

        return variables

    def _clean(self, dict: Union[MutableMapping, None] = None):
        # sourcery skip: avoid-builtin-shadow
        '''
        Finds and removes doublettes of following PLACEHOLDER keys within self.data
        - BLOCKCOMMENT
        - INCLUDE
        - LINECOMMENT
        By definition, only keys on the same nest level are checked for doublettes.
        Doublettes are identified through equality with their lookup values.
        '''
        # START at nest level 0
        if dict is None:
            dict = self.data

        # IDENTIFY all placeholders on current level
        keys_on_this_level = list(dict.keys())
        block_comments_on_this_level = []
        includes_on_this_level = []
        line_comments_on_this_level = []
        for key in keys_on_this_level:
            if re.search(r'BLOCKCOMMENT\d{6}', key):
                block_comments_on_this_level.append(key)
            elif re.search(r'INCLUDE\d{6}', key):
                includes_on_this_level.append(key)
            elif re.search(r'LINECOMMENT\d{6}', key):
                line_comments_on_this_level.append(key)

        unique_block_comments_on_this_level = []                    # BLOCKCOMMENTs
        for key in block_comments_on_this_level:
            with contextlib.suppress(Exception):
                id = int(re.findall(r'\d{6}', key)[0])
                value = str(self.block_comments[id])
                if value in unique_block_comments_on_this_level:    # Found doublette
                    del dict[key]                                   # remove from current level in self.data (the dict)
                    del self.block_comments[id]                     # ..AND from self.block_comments (the lookup table)
                else:                                               # Unique
                    unique_block_comments_on_this_level.append(value)
        unique_includes_on_this_level = []                          # INCLUDEs
        for key in includes_on_this_level:
            with contextlib.suppress(Exception):
                id = int(re.findall(r'\d{6}', key)[0])
                value = self.includes[id]
                if value in unique_includes_on_this_level:          # Found doublette
                    del dict[key]                                   # remove from current level in self.data (the dict)
                    del self.includes[id]                           # ..AND from self.includes (the lookup table)
                else:                                               # Unique
                    unique_includes_on_this_level.append(value)
        unique_line_comments_on_this_level = []                     # LINECOMMENTs
        for key in line_comments_on_this_level:
            with contextlib.suppress(Exception):
                id = int(re.findall(r'\d{6}', key)[0])
                value = self.line_comments[id]
                if value in unique_line_comments_on_this_level:     # Found doublette
                    del dict[key]                                   # remove from current level in self.data (the dict)
                    del self.line_comments[id]                      # ..AND from self.line_comments (the lookup table)
                else:                                               # Unique
                    unique_line_comments_on_this_level.append(value)
                                                                    # RECURSION for nested levels
        for key in dict.keys():
            if isinstance(dict[key], MutableMapping):
                self._clean(dict[key])                              # Recursion

        return


def order_keys(arg: MutableMapping[_KT, _VT]) -> MutableMapping[_KT, _VT]:
    """alpha-numeric sorting of keys, recursively

    Parameters
    ----------
    arg : MutableMapping[_KT, _VT]
        dict, the keys of which shall be sorted

    Returns
    -------
    MutableMapping[_KT, _VT]
        passed in dict with keys sorted
    """
    if isinstance(arg, MutableMapping):
        sorted_dict: MutableMapping[_KT, _VT] = dict(
            sorted(arg.items(), key=lambda x: (isinstance(x[0], str), x[0]))
        )
        for key, value in sorted_dict.items():
            if isinstance(value, dict):
                sorted_dict[key] = order_keys(sorted_dict[key])                                                                # type: ignore
        return sorted_dict
    else:
        logger.warning(
            'dict.order_keys(): no alpha-numeric sorting of keys possible because of \'argument not a dict\', returning same.'
        )
        return arg


def find_global_key(arg: Union[MutableMapping, MutableSequence],
                    query: str = '') -> Union[MutableSequence, None]:
    """Returns the global key thread to the first key the value of which matches the passed in query.

    Parameters
    ----------
    arg : Union[MutableMapping, MutableSequence]
        dict to search in for the queried value
    query : str, optional
        query string for the value to search for, by default ''

    Returns
    -------
    Union[MutableSequence, None]
        global key thread to the first key the value of which matches the passed in query, if found. Otherwise None.
    """
    global_key = []
    if isinstance(arg, MutableMapping):     # dict
        for key, _ in sorted(arg.items()):
            if isinstance(arg[key], (MutableMapping, MutableSequence)):
                if next_level_key := find_global_key(arg=arg[key], query=query):
                    global_key.append(key)
                    global_key.extend(next_level_key)
                    break
            elif re.search(query, str(arg[key])):
                global_key.append(key)
                break
    elif isinstance(arg, MutableSequence):  # list
        for index, _ in enumerate(arg):
            if isinstance(arg[index], (MutableMapping, MutableSequence)):
                if next_level_key := find_global_key(arg=arg[index], query=query):
                    global_key.append(index)
                    global_key.extend(next_level_key)
                    break
            elif re.search(query, str(arg[index])):
                global_key.append(index)
                break
    else:
        logger.warning('Run into not implemented alternative')

    return global_key or None


def set_global_key(arg: MutableMapping, global_key: MutableSequence, value: Any = None):
    """Sets the value for the passed in global key.

    Parameters
    ----------
    arg : MutableMapping
        dict the target key in which shall be set
    global_key : MutableSequence, optional
        list of keys defining the global key thread to the target key (such as returned by method find_global_key())
    value : Any, optional
        value the target key shall be set to, by default None
    """
    if isinstance(global_key, MutableSequence) and (global_key):
        last_branch = arg
        remaining_keys = global_key
        ii = 0
        while len(
            remaining_keys
        ) > 1:                                              # as long as we didn't arrive at the last branch (the one that contains the target key)..
            last_branch = last_branch[remaining_keys[0]]    # ..walk one level further down
            remaining_keys = remaining_keys[1:]
            ii += 1
            if ii == 10: break
        last_branch[remaining_keys[0]] = value

    return


def global_key_exists(arg: MutableMapping, global_key: MutableSequence) -> bool:
    """Checks whether the specified global key exists in the passed in dict.

    Parameters
    ----------
    arg : MutableMapping
        dict to check for existence of the specified global key
    global_key : MutableSequence, optional
        global key the existence of which is checked in the passed in dict

    Returns
    -------
    bool
        True if the specified global key exists, otherwise False
    """
    global_key = global_key or []
    _arg = arg
    for key in global_key:
        try:
            _arg = _arg[key]
        except KeyError:
            return False
    return True


def _merge_dicts(target_dict: MutableMapping, dict_to_merge: MutableMapping, overwrite=False):
    """Merges dict_to_merge into target_dict.

    In contrast to dict.update(), _merge_dicts() works recursively. That is, it does not simply substitute top-level keys
    in target_dict but recursively merges (potentially nested) content from dict_to_merge into target_dict.
    This prevents nested keys from being deleted.

    Parameters
    ----------
    target_dict : MutableMapping
        target dict
    dict_to_merge : MutableMapping
        dict to be merged into target dict
    overwrite : bool, optional
        if True, existing keys will be overwritten, by default False
    """
    for key in dict_to_merge.keys():
        if (key in target_dict.keys()) and isinstance(
            target_dict[key], MutableMapping
        ) and isinstance(dict_to_merge[key], MutableMapping):                                           # dict
            _merge_dicts(target_dict[key], dict_to_merge[key], overwrite)                               # Recursion
        else:
            value_in_target_dict_contains_circular_reference = False
            if key in target_dict and isinstance(target_dict, CppDict):
                value = _insert_expression(target_dict[key], target_dict)
                value_in_target_dict_contains_circular_reference = _value_contains_circular_reference(
                    key, value
                )
            if overwrite or key not in target_dict or value_in_target_dict_contains_circular_reference:
                target_dict[key] = dict_to_merge[key]                                                   # Update

    return


def _insert_expression(value: str, dict: CppDict) -> str:
    if isinstance(value, str) and isinstance(dict,
                                             CppDict) and re.search(r'EXPRESSION\d{6}', value):
        if match_index := re.search(r'\d{6}', value):
            index = int(match_index[0])
            value = dict.expressions[index]['expression'] if index in dict.expressions else value
    return value


def _value_contains_circular_reference(key: str, value: str) -> bool:
    if isinstance(key, str) and isinstance(value, str):
        return key in value
    return False



import re
from difflib import ndiff
import logging

__all__ = ['remove_quotes', 'string_diff']


logger = logging.getLogger(__name__)


def remove_quotes(string):
    '''
    Removes quotes (single or double quotes) from the string object passed in.
    Not only leading and trailing quotes are removed; also any quotes inside a string, if so, are removed.
    '''
    # search_pattern = re.compile(r'(^[\'\\"]{1}|[\'\\"]{1}$)')  #Removes only leading and trailing quotes. Quotes inside a string are kept.
    search_pattern = re.compile(r'[\'\"]')  # Removes ALL quotes in a string. Meaning, not only leading and trailing quotes, but also quotes inside a string are removed.
    # Remove quotes and return
    string = re.sub(search_pattern, '', string)
    return string


def string_diff(line_1, line_2):
    '''
    diff line by line
    printing only diff
    '''
    line_1 = re.split('[\r\n]+', line_1)
    line_2 = re.split('[\r\n]+', line_2)
    diffs = []
    for index, item in enumerate(line for line in ndiff(line_1, line_2) if not re.search(r'^\s*$', line)):
        if re.match(r'^[+\-]', item):
            print('diff in line %4i:' % index, item)
            diffs.append((item))

    line_length = len(diffs)
    print(
        f"There {'is' if line_length == 1 else 'are'} {'no' if line_length == 0 else str(line_length)} difference{'' if line_length == 1 else 's'} between text 1 and text 2."
    )



import logging
import sys
from pathlib import Path
from typing import Union

__all__ = ['configure_logging']

logger = logging.getLogger(__name__)


def configure_logging(
    log_level_console: str = 'WARNING',
    log_file: Union[Path, None] = None,
    log_level_file: str = 'WARNING',
):                                          # sourcery skip: extract-duplicate-method

    log_level_console_numeric = getattr(logging, log_level_console.upper(), None)
    if not isinstance(log_level_console_numeric, int):
        raise ValueError(f'Invalid log level to console: {log_level_console_numeric}')

    log_level_file_numeric = getattr(logging, log_level_file.upper(), None)
    if not isinstance(log_level_file_numeric, int):
        raise ValueError(f'Invalid log level to file: {log_level_file_numeric}')

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(log_level_console_numeric)
    console_formatter = logging.Formatter('%(levelname)-8s %(message)s')
    console_handler.setFormatter(console_formatter)
    root_logger.addHandler(console_handler)

    if log_file:
        if not log_file.parent.exists():
            log_file.parent.mkdir(parents=True, exist_ok=True)
        file_handler = logging.FileHandler(str(log_file.absolute()), 'a')
        file_handler.setLevel(log_level_file_numeric)
        file_formatter = logging.Formatter(
            '%(asctime)s %(levelname)-8s %(message)s', '%Y-%m-%d %H:%M:%S'
        )
        file_handler.setFormatter(file_formatter)
        root_logger.addHandler(file_handler)

    return



import contextlib
import logging
import os
import re
from copy import deepcopy
# Import from math all functions we want to allow inside expressions in a dict.
# This is a bit ugly, but necessary to enable evaluation of parsed expressions with the help of eval().
from math import (  # noqa: F401
    acos, asin, atan, atan2, cos, e, exp, log, log10, pi, pow, sin, sqrt, tan
)
from pathlib import Path
from typing import MutableMapping, MutableSequence, Union

from dictIO import CppDict, Parser
from dictIO.utils.counter import DejaVue


__ALL__ = ['DictReader']

logger = logging.getLogger(__name__)


class DictReader():
    """Reader for dictionaries in dictIO dict file format, as well as JSON and XML
    """

    def __init__(self):
        return

    @staticmethod
    def read(
        source_file: Union[str, os.PathLike[str]],
        includes: bool = True,
        order: bool = False,
        comments: bool = True,
        scope: Union[MutableSequence[str], None] = None,
        parser: Union[Parser, None] = None,
    ) -> CppDict:
        """Reads a dictionary file in dictIO dict file format, as well as JSON and XML.

        Reads a dict file, parses it and transforms its content into a dictIO dict object (CppDict).
        Following file formats are supported and interpreted through source_file's file ending:
        no file ending   ->   dictIO dict file
        '.cpp'           ->   dictIO dict file
        '.foam'          ->   Foam dictionary file
        '.json'          ->   Json dictionary file
        '.xml'           ->   XML file
        Return type is in all cases CppDict

        Parameters
        ----------
        source_file : Union[str, os.PathLike[str]]
            dict file to be read
        includes : bool, optional
            merge sub-dicts being referenced through #include directives, by default True
        order : bool, optional
            sort the read dict, by default False
        comments : bool, optional
            reads comments from source file, by default True
        scope : MutableSequence[str], optional
            scope the dict will be reduced to after reading, by default None
        parser : Parser, optional
            Parser object to be used, by default None

        Returns
        -------
        CppDict
            the read dict

        Raises
        ------
        FileNotFoundError
            if source_file does not exist
        """

        # Make sure source_file argument is of type Path. If not, cast it to Path type.
        source_file = source_file if isinstance(source_file, Path) else Path(source_file)
        if not source_file.exists():
            logger.error(f'source_file not found: {source_file}')
            raise FileNotFoundError(source_file)

        # Create parser
        # If a parser has been passed to read(), use that.
        # Otherwise choose the parser depending on source_file.
        # source_file = Path.joinpath(Path.cwd(), source_file)
        parser = parser or Parser.get_parser(source_file)

        # Parse the dict file and transform it into a CppDict
        parsed_dict = parser.parse_file(source_file, comments=comments)

        # Merge dict files included through #include directives, if not actively refrained through opts
        if includes:
            __class__._merge_includes(parsed_dict, comments=comments)

        # Evaluate and insert back expressions
        __class__._eval_expressions(parsed_dict)

        # Reduce scope of the dict if requested through opts
        if scope:
            # We need here safety hook when specified scope is not found: simply stop (not continue with whole content!)
            if parsed_dict.global_key_exists(scope):
                parsed_dict.reduce_scope(scope)
            else:
                logger.error(f'scope {scope} does not exist in dictionary')
                exit(1)

        # Order the dict, if not actively refrained through opts
        if order:
            parsed_dict.order_keys()

        # Remove includes from the parsed dictionary if requested through opts
        # @TODO: Really necessary?
        # generally a good idea to have a switch suppressing #include ... in output
        # gives the option to disable include (foam required)
        # also to consider: is it really neccessary to have the included dict merged in to current dict.data?
        # if we could avoid that we get a more readable structure, even after some farn operations
        if not includes:
            __class__._remove_include_keys(parsed_dict.data)

        return parsed_dict

    @staticmethod
    def _merge_includes(dict: CppDict, comments: bool = True):
        '''
        Parses and merges any (child) dicts that are referenced in the dict file through #include directives
        '''
        # Create dejavue string watchdog
        djv = DejaVue()
        djv.reset()

        # Inner function: Merge all includes, recursively
        def _merge_includes_recursive(dict: CppDict):

            # empty dict to merge in temporarily, avoiding dict-has-change-error inside the for loop
            temp_dict = CppDict()

            # loop over all possible includes
            for _, _, path in dict.includes.values():
                prove_recursive_include = djv(path)

                if prove_recursive_include is True:
                    call_chain = '->'.join([p.name for p in djv.strings])
                    logger.warning(
                        f'Recursive include detected. Merging of {call_chain} into {dict.name} aborted.'
                    )
                elif not path.exists():
                    logger.warning(f'included dict not found. Merging of {path} aborted.')
                else:
                    parser = Parser.get_parser(path)
                    included_dict = parser.parse_file(path, None, comments=comments)

                    # recursion in case the i-th include also has includes
                    if len(included_dict.includes) != 0:
                        nested_included_dict = _merge_includes_recursive(included_dict)
                        # merge second level
                        temp_dict.merge(nested_included_dict)

                    # merge first level
                    temp_dict.merge(included_dict)

            # merge all in for loop
            dict.merge(temp_dict)

            return dict

        # Call inner funtion to merge all includes, recursively
        dict.merge(_merge_includes_recursive(dict))

        return

    @staticmethod
    def _resolve_reference(ref, vars):
        # resolves a single reference
        ret = None
        try:
            # extract indices, ugly version, nice version is re.sub with a positive lookahead
            indexing = re.findall(r'\[.+\]$', ref)[0]
        except Exception:
            indexing = ''

        ref = re.sub(r'(^\$|\[.+$)', '', ref)   # remove leading $ or trailing [

        if ref in vars:
            ret = vars[ref]     # singular value or field

            ref_changed_through_recursion = False
            while re.search(
                r'\$', str(ret)
            ):                                                  # resolve nested references, if existing, through recursion
                ref = ret
                ref_changed_through_recursion = True
                ret = __class__._resolve_reference(ref, vars)   # recursion
            if ref_changed_through_recursion:
                ref = re.sub(r'(^\$|\[.+$)', '', ref)           # remove leading $ or trailing [
            if indexing:
                with contextlib.suppress(Exception):
                    ret = eval(
                        f"vars['{ref}']{indexing}"
                    )                                           # return the value of the referenced variable (at the specified index, if given)
        return ret

    @staticmethod
    def _eval_expressions(dict: CppDict):
        # Collect all references contained in expressions
        references = []
        for item in dict.expressions.values():
            refs = re.findall(r'\$\w[\w\[\]]+', item['expression'])
            references.extend(refs)
        # Resolve references
        variables = dict.variables
        references = {ref: __class__._resolve_reference(ref, variables) for ref in references}
        references_resolved = {
            ref: value
            for ref,
            value in references.items()
            if (value is not None) and (not re.search(r'EXPRESSION|\$', str(value)))
        }
        references_not_resolved = [ref for ref in references if ref not in references_resolved]

        # Iteratively try to evaluate expressions contained in the dict and then re-resolve all references
        # With every iteration, this should reduce the number of remaining, non resolved references
        references_not_resolved_old = len(references_not_resolved) + 1
        keep_on = True
        while keep_on:
            references_not_resolved_old = len(references_not_resolved)
            expressions_copy = deepcopy(dict.expressions)
            for key, item in expressions_copy.items():
                placeholder = str(item['name'])
                expression = str(item['expression'])
                refs = re.findall(r'\$\w[\w\[\]]+', expression)
                for ref in refs:
                    if ref in references_resolved:
                        expression = re.sub(
                            f'{re.escape(ref)}', str(references_resolved[ref]), expression
                        )

                eval_successful = False
                eval_result = None
                if '$' not in expression:
                    try:
                        eval_result = eval(expression)
                        eval_successful = True
                    except NameError:
                        eval_result = expression
                        eval_successful = True
                    except SyntaxError:
                        logger.warning(
                            'DictReader.(): evaluation of \"%s\" not yet possible' % expression
                        )
                if eval_successful:
                    while global_key := dict.find_global_key(query=placeholder):
                        # Substitute the placeholder in the dict with the result of the evaluated expression
                        dict.set_global_key(global_key, value=eval_result)
                    del dict.expressions[key]
                else:
                    # update the item in dict.expressions with the (at least partly) resolved expression
                    dict.expressions[key]['expression'] = expression

            # At the end of each iteration, re-resolve all references based on the now updated variables table of dict
            references = []
            for item in dict.expressions.values():
                refs = re.findall(r'\$\w[\w\[\]]+', item['expression'])
                references.extend(refs)
            variables = dict.variables
            references = {ref: __class__._resolve_reference(ref, variables) for ref in references}
            references_resolved = {
                ref: value
                for ref,
                value in references.items()
                if (value is not None) and (not re.search(r'EXPRESSION|\$', str(value)))
            }
            references_not_resolved = [ref for ref in references if ref not in references_resolved]

            keep_on = (len(references_not_resolved) < references_not_resolved_old)

        # For expressions that could NOT successfully be evaluated, even after iteration:
        # Back insert the expression string into the dict
        for key, item in dict.expressions.items():
            placeholder = str(item['name'])
            expression = str(item['expression'])
            while global_key := dict.find_global_key(query=placeholder):
                # Substitute the placeholder with the original (or at least partly resolved) expression
                dict.set_global_key(global_key, value=expression)
        dict.expressions.clear()

        return

    @staticmethod
    def _remove_comment_keys(data: MutableMapping):
        '''
        remove comments from data structure for read function call from other programs
        '''
        remove = '[A-Z]+COMMENT[0-9;]+'

        temp_dict = deepcopy(data)

        with contextlib.suppress(Exception):
            for key in temp_dict.keys():
                if isinstance(data[key], MutableMapping):
                    data.update({key: __class__._remove_comment_keys(data[key])})   # recursion
                elif not re.search(remove, key):
                    data.update({key: temp_dict[key]})
        return

    @staticmethod
    def _remove_include_keys(data: MutableMapping):
        '''
        remove includes from data structure for read function call from other programs
        '''
        remove = 'INCLUDE[0-9;]+'

        temp_dict = deepcopy(data)

        with contextlib.suppress(Exception):
            for key in temp_dict.keys():
                if re.search(remove, key):
                    data.pop(key)
        return



import errno
import logging
import os
from pathlib import Path
from typing import List, Sequence, Tuple

__all__ = ['highest_common_root_folder', 'relative_path']

logger = logging.getLogger(__name__)


def highest_common_root_folder(paths: Sequence[Path]) -> Path:
    """Returns the highest common root folder among the passed in paths.

    Parameters
    ----------
    paths : Sequence[Path]
        A sequence of path objects. Can be files or folders, or both.

    Returns
    -------
    Path
        The highest common root folder among the passed in paths.

    Raises
    ------
    ValueError
        If argument 'paths' is empty or if the passed in paths do not share a common root folder.
    """

    if not paths:
        raise ValueError("argument 'paths' is empty.")

    folders: List[Path] = list(paths)
    for path in paths:
        _path: Path = path.resolve()
        _folder: Path = _path.parent if _path.suffix else _path
        folders.append(_folder.absolute())

    if len(folders) == 1:
        return folders[0]

    # Find highest common root folder
    folders_as_parts: List[Tuple[str]] = [
        folder.parts for folder in folders
    ]
    folders_as_parts.sort(key=lambda x: len(x), reverse=False)
    shortest_folder_as_parts: Tuple[str] = folders_as_parts[0]
    common_parts: List[str] = []
    for index, part in enumerate(shortest_folder_as_parts):
        if len({folder_as_parts[index] for folder_as_parts in folders_as_parts}) > 1:
            break
        common_parts.append(part)
    if common_parts:
        return Path(*tuple(common_parts))
    else:
        raise ValueError('The passed in paths do not share a common root folder.')


def relative_path(from_path: Path, to_path: Path) -> Path:
    """Returns the relative path from one path to another.

    Parameters
    ----------
    from_path : Path
        The start point path.
    to_path : Path
        The end point path.

    Returns
    -------
    Path
        The relative path from 'from_path' (the start point) to 'to_path' (the end point).

    Raises
    ------
    ValueError
        If no relative path between 'from_path' and 'to_path' can be resolved.
    """
    relative_path: Path
    try:
        relative_path = to_path.relative_to(from_path)
    except ValueError:
        msg = (
            'Resolving relative path failed using pathlib.\n'
            'Next try will use os.path instead of pathlib.'
        )
        logger.debug(msg)
        try:
            relative_path = Path(
                os.path.relpath(to_path, from_path)
            )
            msg = (
                'Resolving relative path succeeded using os.path'
            )
            logger.debug(msg)
        except Exception as e:
            raise ValueError('Resolving relative path failed using both pathlib and os.path.') from e

    return relative_path



import logging
import os
import re
from pathlib import Path
from typing import Any, MutableMapping, MutableSequence, Tuple, Union
from xml.etree.ElementTree import Element

# from lxml.etree import register_namespace
from lxml.etree import ETCompatXMLParser, fromstring

from dictIO import CppDict
from dictIO.utils.counter import BorgCounter


__ALL__ = ['Parser', 'CppParser', 'FoamParser', 'JsonParser', 'XmlParser']

logger = logging.getLogger(__name__)


class Parser():
    """Base Class for parsers.

    Parsers deserialize a string into a CppDict.
    Subclasses of Parser implement parsing of different, specifically formatted strings (see also Formatters).
    """

    def __init__(self):
        self.counter = BorgCounter()
        self.source_file = None
        return

    @classmethod
    def get_parser(cls, source_file: Union[Path, None] = None):
        """Factory method returning a Parser instance matching the source file type to be parsed

        Parameters
        ----------
        source_file : Path, optional
            name of the source file to be parsed, by default None

        Returns
        -------
        Parser
            specific Parser instance matching the source file type to be parsed
        """
        # Determine the parser to be used by a two stage process:

        # 1. If source_file is passed, choose parser depending on file-ending
        if source_file:
            if source_file.suffix == '.foam':               # .foam -> FoamParser
                return FoamParser()
            elif source_file.suffix == '.json':             # .json -> JsonParser
                return JsonParser()
            elif source_file.suffix in ['.xml', '.ssd']:    # .xml  or  OSP .ssd -> XmlParser
                return XmlParser()

        # 2. If no source file is passed, return CppParser as default / fallback
        return CppParser()  # default

    def parse_file(
        self,
        source_file: Union[str, os.PathLike[str]],
        target_dict: Union[CppDict, None] = None,
        comments: bool = True,
    ) -> CppDict:
        # sourcery skip: inline-immediately-returned-variable
        """Parses a file and deserializes it into a dict.

        Parameters
        ----------
        source_file : Union[str, os.PathLike[str]]
            name of the dict file to be parsed
        target_dict : CppDict, optional
            the target dict the parsed dict file shall be merged into, by default None
        comments : bool, optional
            reads comments from source file, by default True

        Returns
        -------
        CppDict
            the parsed dict

        Raises
        ------
        FileNotFoundError
            if source_file does not exist
        """
        # Make sure source_file argument is of type Path. If not, cast it to Path type.
        source_file = source_file if isinstance(source_file, Path) else Path(source_file)
        if not source_file.exists():
            logger.error(f'source_file not found: {source_file}')
            raise FileNotFoundError(source_file)

        self.source_file = source_file

        # Check whether file to read from exists.
        if not self.source_file.exists():
            logger.warning(
                'Parser.parse_file(): File or path does not exist: \'%s\'. Empty dict will be returned.'
                % source_file
            )
            file_content = ''
        else:
            with self.source_file.open('r') as f:
                file_content = f.read()

        # Create target dict in case no specific target dict was passed in
        if target_dict is None:
            target_dict = CppDict(source_file)
        else:
            target_dict.source_file = source_file.absolute()
            target_dict.path = source_file.parent
            target_dict.name = source_file.name

        # one final check only
        # whether a file exists (can also have zero content)
        # or a dict was given (can also contain nothing)
        if not self.source_file.exists() and target_dict is None:
            logger.error(
                'Parser.parse_file(): File or path does not exist (\'%s\') or no target dict (%s) was given.'
                % (source_file, target_dict)
            )

        # Parse file content
        parsed_dict = self.parse_string(file_content, target_dict, comments)

        return parsed_dict

    def parse_string(
        self,
        string: str,
        target_dict: CppDict,
        comments: bool = True,
    ) -> CppDict:
        # sourcery skip: inline-immediately-returned-variable, lift-return-into-if
        """Parses a string and deserializes it into a CppDict.

        Note: Override this method when implementing a specific Parser.

        Parameters
        ----------
        string : str
            the string to be parsed (i.e. the content of the file that had been read using parse_file())
        target_dict : CppDict
            the target dict the parsed dict file shall be merged into
        comments : bool, optional
            reads comments, by default True

        Returns
        -------
        CppDict
            the parsed dict
        """

        # +++VERIFY STRING CONTENT++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Check that string is not empty.
        if not string:
            logger.warning(f'Parser.parse_string(): String to parse is empty: {string}')

        # Create target dict in case no specific target dict was passed in
        if target_dict is None:
            logger.warning(
                'Parser.parse_string(): Target dict is None. Will create new target dict, however, with empty filename.'
            )
            target_dict = CppDict()

        # Create a local CppDict instance where the stringcontent is temporarily parsed into
        if target_dict.source_file:
            parsed_dict = CppDict(target_dict.source_file)
        else:
            parsed_dict = target_dict

        # +++PARSE DICTIONARY+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Specific Parser implementations need to implement this part.

        # +++MERGE PARSED DICTIONARY INTO TARGET DICTIONARY+++++++++++++++++++++++++++++++++++++++++

        # Specific Parser implementations need to, finally, update the target dict!
        # (in the base class, however, this does not make sense - hence commented out)
        # target_dict.merge(parsed_dict)

        # +++RETURN PARSED DICTIONARY+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        return parsed_dict

    def parse_type(self, arg: Any) -> Any:
        """Parses a single value

        Parses a single value and casts it to its native type (str, int, float, boolean or None).

        Parameters
        ----------
        arg : Any
            the value to be parsed

        Returns
        -------
        Any
            the value casted to its native type (str, int, float, boolean or None)
        """

        # Numbers (int and float) are returned without conversion
        if isinstance(arg, int) and not isinstance(arg, bool):  # int
            return arg
        if isinstance(arg, float):                              # float
            return arg

        # Boolean and None types are returned without conversion
        if isinstance(arg, bool):   # bool
            return arg
        if arg is None:             # None
            return arg

        # Any other non-string type: Return without conversion
        if not isinstance(arg, str):
            return arg

        # String: Convert the string content to its native type, where possible.
        # The following starts a filter cascade, hence the sequence is important,
        # otherwise latter clauses would produce unreasonable matches!

        # Empty string
        check: str = __class__.remove_quotes_from_string(arg)
        if not check:
            return ''

        # Simple placeholder or reserved expressions -> do nothing
        # Note: This if clause is important: It avoids that distinct placeholders like e.g.
        # '-' are interpreted as float (and then transformed to float .. what might even fail).
        if arg in ['-', '_', '.']:
            return arg

        # String numbers shall be converted to numbers (int and float)
        if re.search(r'^[+-]?\d+$', arg):                           # int
            return int(arg)
        if re.search(r'^[+-]?(\d+(\.\d*)?|\.\d+)$', arg):           # float
            return float(arg)
        if re.search(r'^[+-]?\d*(\.\d*)?([eE]?[-+]?\d+)?$', arg):   # float written as fpn like 1.e-03
            return float(arg)

        # Booleans and None types that are masked as strings
        # ('True', 'true', 'False', 'false', 'ON', 'on', 'OFF', 'off', 'None', 'none', 'NULL', 'null')
        # shall be converted to its native Boolean or None type, respectively
        if re.search(r'^(true)$', arg.strip().lower()):     # True
            return True
        if re.search(r'^(false)$', arg.strip().lower()):    # False
            return False
        if re.search(r'^(on)$', arg.strip().lower()):       # OpenFOAM 'on' -> True
            return True
        if re.search(r'^(off)$', arg.strip().lower()):      # OpenFOAM 'off' -> False
            return False
        if re.search(r'^(none)$', arg.strip().lower()):     # None
            return None
        if re.search(r'^(null)$', arg.strip().lower()):     # C++ 'NULL' or JSON 'null' -> None
            return None

        # Any other string: return 'as is', but make sure extra quotes, if so, are stripped.
        # Note: Also any placeholder strings will fall into this category.
        # Returned 'as is' they are kept unchanged, what is in fact what we want here.
        return __class__.remove_quotes_from_string(arg)

    def parse_types(
        self, arg: Union[MutableMapping, MutableSequence]
    ) -> Union[MutableMapping, MutableSequence]:
        """Parses multiple values

        Parses all values inside a dict or list and casts them to its native types (str, int, float, boolean or None).
        The function traverses the passed in dict or list recursively
        so that all values in also nested dicts and lists are parsed.

        Parameters
        ----------
        arg : Union[MutableMapping, MutableSequence]
            the dict or list containing the values to be parsed

        Returns
        -------
        Union[MutableMapping, MutableSequence]
            the original dict or list, yet with all contained values being casted to its native types (str, int, float, boolean or None).
        """
        if isinstance(arg, MutableSequence):
            for index, _ in enumerate(arg):
                if isinstance(arg[index], (MutableMapping, MutableSequence)):
                    self.parse_types(arg[index])
                else:
                    arg[index] = self.parse_type(arg[index])
        elif isinstance(arg, MutableMapping):
            for key in arg.keys():
                if isinstance(arg[key], (MutableMapping, MutableSequence)):
                    self.parse_types(arg[key])
                else:
                    arg[key] = self.parse_type(arg[key])
        return arg

    @staticmethod
    def remove_quotes_from_string(arg: str, all: bool = False) -> str:
        """Removes quotes from a string

        Removes quotes (single and double quotes) from the string object passed in.

        Parameters
        ----------
        arg : str
            the string with quotes
        all : bool, optional
            if true, all quotes inside the string will be removed (not only leading and trailing quotes), by default False

        Returns
        -------
        str
            the string with quotes being removed

        Raises
        ------
        TypeError
            if arg is not of type str
        """

        if not isinstance(arg, str):                                                                    # any other type
            raise TypeError(
                f'{__class__.__name__}.remove_quotes_from_string(): invalid type of paramter arg:\n'
                f'arg is of type {type(arg)!s} but expected to be str.'
            )

        if all:
            # Removes ALL quotes in a string:
            # Not only leading and trailing quotes, but also quotes inside a string are removed.
            search_pattern = re.compile(r'[\'\"]')
        else:
            # Removes only leading and trailing quotes. Quotes inside a string are kept.
            search_pattern = re.compile(r'(^[\'\\"]{1}|[\'\\"]{1}$)')

        return re.sub(search_pattern, '', arg)

    @staticmethod
    def remove_quotes_from_strings(
        arg: Union[MutableMapping, MutableSequence]
    ) -> Union[MutableMapping, MutableSequence]:
        """Removes quotes from multiple strings

        Removes quotes (single and double quotes) from all string objects inside a dict or list.
        The function traverses the passed in dict or list recursively
        so that all strings in also nested dicts and lists are processed.


        Parameters
        ----------
        arg : Union[MutableMapping, MutableSequence]
            the dict or list containing strings the quotes in which shall be removed

        Returns
        -------
        Union[MutableMapping, MutableSequence]
            the original dict or list, yet with quotes in all strings being removed

        Raises
        ------
        TypeError
            if arg is not of type MutableMapping or MutableSequence (u.e. dict or list)
        """
        if not isinstance(arg, (MutableMapping, MutableSequence)):                                      # any other type
            raise TypeError(
                f'{__class__.__name__}.remove_quotes_from_strings(): invalid type of paramter arg:\n'
                f'arg is of type {type(arg)!s} but expected to be MutableMapping | MutableSequence.'
            )

        if isinstance(arg, MutableMapping):                                     # dict
            for key in arg.keys():
                if isinstance(arg[key], (MutableMapping, MutableSequence)):     # dict or list
                    arg[key] = __class__.remove_quotes_from_strings(arg[key])   # (recursion)
                elif isinstance(arg[key], str):                                 # str
                    arg[key] = __class__.remove_quotes_from_string(arg[key])
        elif isinstance(arg, MutableSequence):                                  # list
            for i in range(len(arg)):
                if isinstance(arg[i], (MutableMapping, MutableSequence)):       # dict or list
                    arg[i] = __class__.remove_quotes_from_strings(arg[i])       # (recursion)
                elif isinstance(arg[i], str):                                   # str
                    arg[i] = __class__.remove_quotes_from_string(arg[i])

        return arg


class CppParser(Parser):
    """Parser to deserialize a string in dictIO dict file format into a CppDict.
    """

    def __init__(self):
        '''
        Implementation specific default configuration of CppParser
        '''
        # Invoke base class constructor
        super().__init__()

    def parse_string(
        self,
        string: str,
        target_dict: CppDict,
        comments: bool = True,
    ) -> CppDict:
        """Parses a string in dictIO dict file format and deserializes it into a CppDict.

        Parameters
        ----------
        string : str
            the string to be parsed (i.e. the content of the file that had been read using parse_file())
        target_dict : CppDict
            the target dict the parsed dict file shall be merged into
        comments : bool, optional
            reads comments, by default True

        Returns
        -------
        CppDict
            the parsed dict
        """

        # +++CALL BASE CLASS IMPLEMENTATION+++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        parsed_dict = super().parse_string(string, target_dict)

        # +++PARSE LINE CONTENT+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Split file content into lines and store them in the newly created CppDict instance
        parsed_dict.line_content = string.splitlines(keepends=True)

        # Extract line comments
        self._extract_line_comments(parsed_dict, comments)

        # Extract include directives
        self._extract_includes(parsed_dict)

        # +++PARSE BLOCK CONTENT++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Concatenate all lines from line_content
        # As extracting block comments is easier with line endings still existing, at first we preserve them.
        self._convert_line_content_to_block_content(parsed_dict)    # preserves line endings

        # Extract block comments      ..and remove line endings right thereafter

        self._extract_block_comments(parsed_dict, comments)
        self._remove_line_endings_from_block_content(parsed_dict)

        # Extract string literals
        self._extract_string_literals(parsed_dict)

        # Extract expressions
        self._extract_expressions(parsed_dict)

        # Make sure that all delimiters are surrounded by at least one space before and after
        # to ensure they are properly identified when we tokenize block_content
        self._separate_delimiters(parsed_dict)

        # +++PARSE TOKENS+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Split block_content into tokens
        self._convert_block_content_to_tokens(parsed_dict)

        # Determine the hierarchic level of each token and assign it to the token (the hierarchy is dictated by the sequence of delimiters)
        self._determine_token_hierarchy(parsed_dict)

        # Parse the hierarchic tokens
        self._convert_tokens_to_dict(parsed_dict)

        # Insert back string literals
        self._insert_string_literals(parsed_dict)

        # +++CLEAN PARSED DICTIONARY++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        self._clean(parsed_dict)

        # +++MERGE PARSED DICTIONARY INTO TARGET DICTIONARY+++++++++++++++++++++++++++++++++++++++++
        target_dict.merge(parsed_dict)

        # +++RETURN PARSED DICTIONARY+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        return parsed_dict

    def _extract_line_comments(self, dict: CppDict, comments: bool):
        """Finds and extracts C++ line comments (// ..) from dict.line_content, and replaces them with Placeholders.

        Finds C++ line comments (// line_comment), extracts them,
        and replaces the complete line with a placeholder in the form LINECOMMENT000000 .
        The extracted line comments are stored in .line_comments as key value pairs {index:line_comment}.
        index, therein, corresponds to the integer number in LINECOMMENT000000.

        Parameters
        ----------
        dict : CppDict
            dict to be processed. _extract_line_comments() works on dict.line_content
        comments : bool
            If False, line comments will be removed (they get replaced by an empty placeholder then, which in effect removes them).
        """
        for index, line in enumerate(dict.line_content):
            # if it is a line comment or just a "http://"?
            if re.search(r'(?<!:)/{2}.*$', line):
                # if re.search(r'/{2}.*$', line):
                key = self.counter()
                # Search for only the FIRST occurrence of '//' in the line.
                # From there, consider all chars until line ending as ONE comment.
                line_comment = re.findall('/{2}.*$', line)[0]
                dict.line_comments.update({key: line_comment})
                placeholder = 'LINECOMMENT%06i' % key
                if not comments:
                    placeholder = ''
                # Replace line comment with placeholder
                dict.line_content[index] = dict.line_content[index].replace(
                    line_comment, placeholder
                )

        return

    def _extract_includes(self, dict: CppDict):
        """Finds and extracts #include directives from dict.line_content, and replaces them with Placeholders.

        Finds #includes directives (#include file), extracts them,
        and replaces the complete line where the include directive was found
        with a placeholder in the form #INCLUDE000000.
        The absolute path to the file referenced in the include directive is determined.
        The original line with its include directive as well as the absolute path to the file to include
        is then stored as a key-value pair in dict.includes, in the form {index:(include_directive, include_file_name, include_file_path)}
        index, therein, corresponds to the integer number in #INCLUDE000000.

        Parameters
        ----------
        dict : CppDict
            dict to be processed. _extract_includes() works on dict.line_content
        """

        for index, line in enumerate(dict.line_content):
            if re.search(r'^\s*#\s*include', line):
                ii = self.counter()
                dict.line_content[index] = 'INCLUDE%06i\n' % ii

                include_file_name = re.sub(r'(^\s*#\s*include\s*|\s*$)', '', line)
                include_file_name = __class__.remove_quotes_from_string(include_file_name)

                include_file_path = Path.joinpath(dict.path, include_file_name)

                include_directive = line
                if line[-1] == '\n':
                    include_directive = line[:-1]

                dict.includes.update(
                    {ii: (include_directive, include_file_name, include_file_path)}
                )

        return

    def _convert_line_content_to_block_content(self, dict: CppDict):
        """Concatenates all lines from line_content to one long string (text block) and stores the result in block_content.
        """
        dict.block_content = ''.join(dict.line_content)
        dict.line_content.clear()
        return

    def _remove_line_endings_from_block_content(self, dict: CppDict):
        """Removes all line endings in .block_content and substuitutes them by single spaces.
        """
        dict.block_content = re.sub(r'\n', ' ', dict.block_content).strip()
        return

    def _extract_block_comments(self, dict: CppDict, comments: bool):
        """Finds and extracts C++ block comments (/* .. */) from dict.block_content, and replaces them with Placeholders.

        Finds C++ block comments (/* block_comment */), extracts them,
        and replaces them with a placeholder in the form BLOCKCOMMENT000000.
        The extracted block comments are stored in .block_comments as key value pairs {index:block_comment}.
        index, therein, corresponds to the integer number in BLOCKCOMMENT000000.

        Parameters
        ----------
        dict : CppDict
            dict to be processed. _extract_block_comments() works on dict.block_content
        comments : bool
            If False, block comments will be removed (they get replaced by an empty placeholder then, which in effect removes them).
        """

        block_comments = re.findall(r'/\*[\w\W\d\D\s]*?\*/', dict.block_content, re.MULTILINE)
        dict.block_comments = {i: block_comments[i] for i in range(len(block_comments))}

        for key, block_comment in dict.block_comments.items():
            placeholder = 'BLOCKCOMMENT%06i' % key
            # placeholder = 'BLOCKCOMMENT%06i BLOCKCOMMENT%06i;' % (key, key)
            if not comments:
                placeholder = ''
            # Replace block comment with placeholder
            dict.block_content = re.sub(re.escape(block_comment), placeholder, dict.block_content)

        return

    def _extract_string_literals(self, dict: CppDict):
        """Finds and extracts string literals from dict.block_content, and replaces them with Placeholders.

        Finds string literals, extracts them,
        and replaces them with a placeholder in the form STRINGLITERAL000000.
        Substrings within .block_content that are surrounded by single quotes are identified as string literals.
        The extracted string literals are stored in .string_literals as key value pairs {index:string_literal}.
        index, therein, corresponds to the integer number in STRINGLITERAL000000.

        Parameters
        ----------
        dict : CppDict
            dict to be processed. _extract_string_literals() works on dict.block_content.
        """

        # Step 1: Find single quoted string literals in .block_content
        search_pattern = r'\'.*?\''
        string_literals = re.findall(search_pattern, dict.block_content, re.MULTILINE)
        for string_literal in string_literals:

            index = self.counter()
            placeholder = 'STRINGLITERAL%06i' % index

            # Replace all occurances of the string literal in .block_content with the placeholder (STRINGLITERAL000000)
            # Note: For re.sub() to work properly we need to escape all special characters
            search_pattern = re.compile(re.escape(string_literal))
            dict.block_content = re.sub(
                search_pattern, placeholder, dict.block_content
            )                                                       # replace string literal in .block_content with placeholder

            # Register the string literal in .string_literals
            dict.string_literals.update(
                {index: __class__.remove_quotes_from_string(string_literal)}
            )

        # Step 2: Find double quoted string literals in .block_content
        # Double quoted strings are identified as string literals only in case they do not contain a $ character.
        # (double quoted strings containing a $ character are considered expressions, not string literals.)
        search_pattern = r'\".*?\"'
        string_literals = re.findall(search_pattern, dict.block_content, re.MULTILINE)
        for string_literal in string_literals:
            if '$' not in string_literal:

                index = self.counter()
                placeholder = 'STRINGLITERAL%06i' % index

                # Replace all occurances of the string literal in .block_content with the placeholder (STRINGLITERAL000000)
                # Note: For re.sub() to work properly we need to escape all special characters
                search_pattern = re.compile(re.escape(string_literal))
                dict.block_content = re.sub(
                    search_pattern, placeholder, dict.block_content
                )                                                       # replace expression in .block_content with placeholder

                # Register the string literal in .string_literals
                dict.string_literals.update(
                    {index: __class__.remove_quotes_from_string(string_literal)}
                )

        return

    def _extract_expressions(self, dict: CppDict):
        """Finds and extracts expressions from dict.block_content, and replaces them with Placeholders.

        Finds expressions, extracts them,
        and replaces them with a placeholder in the form EXPRESSION000000.
        Substrings within .block_content that are surrounded by double quotes and contain minimum one $reference are identified as expressions.
        The extracted expressions are stored in .expressions as a key'd subdict with multiple elements {index:{'index': index, 'expression': expression, 'name': placeholder}}.
        index, therein, corresponds to the integer number in EXPRESSION000000.

        Parameters
        ----------
        dict : CppDict
            dict to be processed. _extract_expressions() works on dict.block_content.
        """

        dict.expressions = {}

        # Step 1: Find expressions in .block_content .
        # Expressions are double quoted strings that contain minimum one reference.
        # References are denoted using the '$' syntax familiar from shell programming.
        # Any key'd entries in a dict are considered variables and can be referenced.
        search_pattern = r'"[^"]*\$.*?"'
        expressions = re.findall(search_pattern, dict.block_content, re.MULTILINE)
        for expression in expressions:

            index = self.counter()
            placeholder = 'EXPRESSION%06i' % index

            # Replace all occurances of the expression in .block_content with the placeholder (EXPRESSION000000)
            # Note: For re.sub() to work properly we need to escape all special characters
            # (covering both '$' as well as any mathematical operators in the expression)
            search_pattern = re.compile(re.escape(expression))
            dict.block_content = re.sub(
                search_pattern, placeholder, dict.block_content
            )                                                       # replace expression in .block_content with placeholder

            # Register the expression in .expressions
            expression = re.sub(r'\"', '', expression)
            dict.expressions.update(
                {index: {
                    'index': index, 'expression': expression, 'name': placeholder
                }}
            )

        # Step 2: Find references in .block_content (single references to key'd entries that are NOT in double quotes).
        search_pattern = r'\$\w[\w\[\]]+'
        while match := re.search(search_pattern, dict.block_content, re.MULTILINE):
            reference = match[0]
            index = self.counter()
            placeholder = 'EXPRESSION%06i' % index
            # Replace the found reference in .block_content with the placeholder (EXPRESSION000000)
            dict.block_content = match.re.sub(placeholder, dict.block_content, count=1)
            # Register the reference as expression in .expressions
            dict.expressions.update(
                {index: {
                    'index': index, 'expression': reference, 'name': placeholder
                }}
            )
        return

    def _separate_delimiters(self, dict: CppDict, delimiters=None):
        """Ensures that delimiters are separated by exactly one space before and after.

        Parses .block_content for occurences of the delimiters passed in, and strips any spaces surrounding each
        delimiter to exactly one single space before and one single space after the delimiter.
        Further, it removes all line endings from .block_content and eventually replaces them with single spaces.

        After _separate_delimiters() returns, .block_content contains only
        - words (with single char delimiters also considered a 'word' here)    and
        - single spaces

        Hence, calling _separate_delimiters() is a preparatory step before
        decomposing .block_content into a list of tokens with re.split('\\s').
        It ensures that re.split('\\s') generates tokens containing one single word each (or a single char delimiter)
        but not any 'waste' tokens with spaces, tabs or line endings will be deleted.
        """

        if delimiters is None:
            delimiters = dict.delimiters

        # Insert at least one \s around every char in list
        for char in delimiters:
            dict.block_content = re.sub(str(r'(\%s)' % char), str(f' {char} '), dict.block_content)

        # Substitute all \s+ to \s
        # This turns multiple spaces into one single space.
        # However, as \s+ matches ANY whitespace character (\s+ is equivalent to [ \t\n\r\f\v]+)
        # this also deletes all line endings (\n). As explained above, this is well intended though.
        dict.block_content = re.sub(r'\s+', ' ', dict.block_content)

        return

    def _convert_block_content_to_tokens(self, dict: CppDict):
        """Decomposes .block_content into a list of tokens.
        """

        dict.tokens = [(0, i) for i in re.split(r'\s', dict.block_content)]
        dict.block_content = ''

        return

    def _determine_token_hierarchy(self, dict: CppDict):
        # sourcery skip: use-join
        """Creates the hierarchy among the tokens and tests their indentation
        """
        level = 0
        count_open = []
        count_close = []
        for index, item in enumerate(dict.tokens):

            if item[1] in dict.openingBrackets:
                push_pop = 1
                count_open.append(item[1])
            elif item[1] in dict.closingBrackets:
                push_pop = -1
                count_close.append(item[1])
            else:
                push_pop = 0

            if push_pop < 0:
                level += push_pop
                push_pop = 0

            # if not re.search('COMMENT', self.tokens[index][1]):
            dict.tokens[index] = (level, dict.tokens[index][1])

            if push_pop > 0:
                level += push_pop
                push_pop = 0

        if level != 0:
            counted = ''
            for opening_bracket, closing_bracket in zip(dict.openingBrackets, dict.closingBrackets):
                counted += ''.join(
                    [
                        '\t\t\t',
                        opening_bracket,
                        str(len([b for b in count_open if b == opening_bracket])),
                        ' -- ',
                        str(len([b for b in count_close if b == closing_bracket])),
                        closing_bracket,
                        '\n'
                    ]
                )
            logger.error(
                '_determine_token_hierarchy: opening and closing delimiters in dict %s are not balanced:\n%s'
                % (dict.name, counted)
            )

        return

    def _convert_tokens_to_dict(self, dict: CppDict):
        """Converts the hierarchic tokens into a dict
        """
        dict.update(self._parse_tokenized_dict(dict))
        dict.tokens.clear()

        return

    def _parse_tokenized_dict(
        self,
        dict: CppDict,
        tokens: Union[MutableSequence[Tuple[int, str]], None] = None,
        level: int = 0,
    ) -> dict:
        """Parses a tokenized dict and returns the parsed dict.

        Parses all tokens, identifies the element within the tokenized dict each token represents or belongs to,
        converts related tokens into the element's type and stores it in local dict (parsed_dict).

        Following elements within the tokenized dict are identified and parsed:
        - nested data struct (list or dict)
        - key value pair
        - comment (string literal containing 'COMMENT')

        After all tokens have successfully been parsed, return the parsed dict.

        Note: To allow recursive calls in case of nested dicts, parsed_dict is declared as a local variable.
        """
        # sourcery skip: remove-redundant-pass

        parsed_dict = {}

        if tokens is None:
            tokens = dict.tokens

        # Iterate through tokens
        token_index = 0
        while token_index < len(tokens):
            # logger.info('token: %s%s' % ('\t'*tokens[tIndex][0], tokens[tIndex][1]))  # 1

            # Nested data struct (list or dict)   '(' = list    '{' = dict
            if tokens[token_index][1] in dict.openingBrackets:

                # The name (key) of the data struct is by convention directly preceeding the opening bracket.
                # ..except if there are line comments in between. skip those:
                offset = 1
                while re.match('^.*COMMENT.*$', str(tokens[token_index - offset][1])):
                    offset += 1
                # name (key) of the data struct:
                name = tokens[token_index - offset][1]

                # Closing bracket has by definition same level as opening bracket.
                # (Note: the tokens BETWEEN the brackets are considered one level 'deeper'; but that's not the point here)
                closing_bracket = self._find_companion(dict, tokens[token_index][1])
                closing_level = tokens[token_index][0]

                # Create a temporary data_struct_tokens list for just the nested data struct, containing
                # all tokens from the opening bracket (first token) to the closing bracket (last token)
                data_struct_tokens = []
                i = 0
                last_index = None
                # Start at opening bracket, go forward and copy the tokens
                # until (and including) the accompanied closing bracket.
                while (
                    tokens[token_index + i][1] != closing_bracket
                    or tokens[token_index + i][0] != closing_level
                    and not re.match('^.*COMMENT.*$', str(tokens[token_index + i][1]))
                ):
                    last_index = token_index + i
                    data_struct_tokens.append(tokens[token_index + i])
                    i += 1
                data_struct_tokens.append(tokens[token_index + i])

                # Do a Syntax-Check at the closing bracket of the data struct.
                # As the syntax for lists and dicts is different, the syntax check is type specific:
                # list:
                # Proof that list properly ends with ';'
                # (= assert that closing bracket of the list is followed by ';')
                if data_struct_tokens[-1][1] == ')' and tokens[token_index + i
                                                               + 1][1] not in [';', ')']:
                    # log error: Missing ';' after list
                    logger.warning(
                        'mis-spelled expression / missing \';\' around \"%s\"' % ' '.join(
                            [name] + [t[1] for t in data_struct_tokens]
                            + [tokens[token_index + i + 1][1]]
                        )
                    )
                # dict:
                if data_struct_tokens[-1][1] == '}':
                    # Proof that last key value pair in dict ends with ';'
                    # (= assert that second-to-last token is ';')
                    index = -2
                    # ..ok, line comments do not count .. skip them:
                    while re.match('^.*COMMENT.*$', str(data_struct_tokens[index][1])):
                        index -= 1
                    # ..but now: Does the last key value pair end with ';'?
                    if data_struct_tokens[index][1] not in ['{', ';', '}']:
                        # log error: Missing ';' after key value pair
                        logger.error(
                            'mis-spelled expression / missing \';\' around \"%s\"'
                            % ' '.join([name] + [t[1] for t in data_struct_tokens])
                        )

                # Parse the tokenized data struct, translate it into its type (list or dict),
                # and update parsed_dict with the new list or dict.
                # Again, the code is type specific depending on whether the parsed data struct is a list or a dict.
                # list:
                if data_struct_tokens[0][1] == '(':
                    # Check whether the list is empty
                    if len(data_struct_tokens) < 3:
                        # is empty (contains only opening and closing bracket)
                        # update parsed_dict with just the empty list
                        parsed_dict[name] = []
                    else:
                        # has content
                        # parse the nested list
                        nested_list = self._parse_tokenized_list(
                            dict, data_struct_tokens, level=level + 1
                        )
                        # update parsed_dict with the nested list
                        parsed_dict[name] = nested_list

                #  dict:
                elif data_struct_tokens[0][1] == '{':
                    # parse the nested dict (recursion)
                    nested_dict = self._parse_tokenized_dict(
                        dict, data_struct_tokens[1:-1], level=level + 1
                    )
                    # update parsed_dict with the nested dict
                    parsed_dict[name] = nested_dict

                # All done: Identified data struct is parsed, translated into its corresponding type,
                # and local parsed_dict is updated.
                # To close out and move on, fast-forward the index of tokens
                # to 'after' the data struct we just parsed:
                if last_index is not None:
                    token_index = last_index + 1

            # Key value pair
            elif tokens[token_index][1] == ';' and tokens[token_index - 1][1] != ')':

                # Read the name (key) and the value from the key value pair
                # Parse from right to left, starting at the identified ';'
                # and then copy the tokens into a temporary key_value_pair_tokens list:
                key_value_pair_tokens: MutableSequence[Tuple[int,
                                                             str]] = [tokens[token_index]]  # ';'
                                                                                            # key_value_pair_tokens.append(tokens[token_index])                       # ';'
                key_value_pair_token_level: int = tokens[token_index][0]
                i = 1
                last_index = None
                while (
                    token_index - i >= 0
                    and tokens[token_index - i][0] == key_value_pair_token_level
                    and tokens[token_index - i][1] not in [';', '}']
                    and not re.match(r'^.*COMMENT.*$', str(tokens[token_index - i][1]))
                    and not re.match(r'^.*INCLUDE.*$', str(tokens[token_index - i][1]))
                ):
                    key_value_pair_tokens.append(tokens[token_index - i])
                    i += 1

                # reverse token list
                key_value_pair_tokens = key_value_pair_tokens[::-1]

                # Before proceeding with reading key and value, make sure that what was parsed
                # really represents a key-value pair, consisting of three elements key, value and ';'
                if not (
                    len(key_value_pair_tokens) == 3 and len(key_value_pair_tokens[0]) == 2
                    and len(key_value_pair_tokens[1]) == 2
                ):
                    # Something is lexically wrong.  Not a valid key-value pair. -> Skip and log warning
                    context_tokens_index_from = max(0, token_index - 20)
                    context_tokens_index_to = min(token_index + 20, len(tokens))
                    context = '/' + ' '.join(
                        [
                            tokens[i][1]
                            for i in range(context_tokens_index_from, context_tokens_index_to)
                            if len(tokens[i]) > 1
                        ]
                    ) + '/'
                    logger.warning(
                        f'CppParser._parse_tokenized_dict(): tokens skipped: {key_value_pair_tokens} inside {context}'
                    )
                else:
                    if len(key_value_pair_tokens) > 3:
                        logger.warning(
                            'CppParser._parse_tokenized_dict(): '
                            'more tokens in key-value pair than expected: %s' %
                            (str(key_value_pair_tokens))
                        )
                    # read the name (key) (first token, by convention)
                    name = key_value_pair_tokens[0][1]
                    # read the value (second token, by convention)
                    value = self.parse_type(key_value_pair_tokens[1][1])
                    # update parsed_dict with the parsed key value pair
                    # Note: Following update would be greedy, if parsed_dict would be declared as global variable.
                    # This exactly is why parsed_dict is declared as local variable in _parse_tokenized_dict().
                    # Doing so, an empty local dict is created with each call to _parse_tokenized_dict(),
                    # and that is, also with each RECURSIVE call.
                    # As every recursive call passes in a temporary token list containing only the nested
                    # data struct, updating a key effects the current (and local) parsed_dict only.
                    # Every key hence is being updated exclusively within its own local context; ambiguous occurences of keys are avoided.
                    if isinstance(name, int):
                        logger.error(f"unexpected type of key 'name': int (value: {name}).")
                    parsed_dict[name] = value

            # Comment
            # As comments cannot be stored "inline" in a structured dictionary
            # the same way as they were stored inline in the dict file parsed,
            # they are stored in parsed_dict in the form of distinct comment keys.
            # As the sequence of keys added to parsed_dict exactly follows the sequence of the tokens parsed,
            # the location of comments in relation to their preceding tokens is preserved.
            # The original comments are back-inserted later in insert_block_comments() and insert_line_comments(), respectively.
            # For now, create an entry with the placeholder string of the comment assigned to both, key and value.
            elif re.match('^.*COMMENT.*$', str(tokens[token_index][1])):
                parsed_dict[tokens[token_index][1]] = tokens[token_index][1]

            # Include directive
            # Create an entry with the placeholder string of the include directive assigned to both, key and value.
            elif re.match('^.*INCLUDE.*$', str(tokens[token_index][1])):
                parsed_dict[tokens[token_index][1]] = tokens[token_index][1]

            # -(Unknown element)
            else:
                pass

            # Iterate to next token
            token_index += 1

        # Return the parsed dict
        return parsed_dict

    def _parse_tokenized_list(
        self,
        dict: CppDict,
        tokens: Union[MutableSequence[Tuple[int, str]], None] = None,
        level: int = 0,
    ) -> list:
        """Parses a tokenized list and returns the parsed list.

        Parses all tokens, identifies the element within the tokenized list each token represents or belongs to,
        converts related tokens into the element's type and stores it in local list (parsed_list).

        Following elements within the tokenized list are identified and parsed:
        - nested data struct (list or dict)
        - single value type

        After all tokens have successfully been parsed, return the parsed list.

        Note: To allow recursive calls in case of nested lists, parsed_list is declared as a local variable.
        """
        # sourcery skip: remove-empty-nested-block, remove-redundant-if, remove-redundant-pass

        parsed_list = []

        if tokens is None:
            tokens = dict.tokens

        # Iterate through tokens
        base_level = tokens[0][0]
        token_index = 0
        while token_index < len(tokens):
            # logger.info('token: %s%s' % ('\t'*tokens[tIndex][0], tokens[tIndex][1]))  # 1

            # Nested data struct (list or dict)   '(' = list    '{' = dict
            if (
                tokens[token_index][1] in dict.openingBrackets
                and tokens[token_index][0] > base_level
            ):
                # Closing bracket has by definition same level as opening bracket.
                # (Note: the tokens BETWEEN the brackets are considered one level 'deeper'; but that's not the point here)
                closing_bracket = self._find_companion(dict, tokens[token_index][1])
                closing_level = tokens[token_index][0]

                # Create a temporary token list for just the nested data struct, containing
                # all tokens from the opening bracket (first token) to the closing bracket (last token)
                temp_tokens = []
                i = 0
                last_index = None
                # Start at opening bracket, go forward and copy the tokens
                # until (and including) the accompanied closing bracket
                while (
                    tokens[token_index + i][1] != closing_bracket
                    or tokens[token_index + i][0] != closing_level
                    and not re.match('^.*COMMENT.*$', str(tokens[token_index + i][1]))
                ):
                    last_index = token_index + i
                    temp_tokens.append(tokens[token_index + i])
                    i += 1
                temp_tokens.append(tokens[token_index + i])

                # Do a Syntax-Check at the closing bracket of the data struct.
                # As the syntax for lists and dicts is different, the syntax check is type specific:
                # list:
                if temp_tokens[-1][1] == ')':
                    # nothing to proof.  A list nested inside a list simply ends with ')'.
                    # Note: This is different for a list nested inside a dict: Then, the closing
                    # bracket of the list must be followed by ';' (because, basically, the list is then the 'value'
                    # part of a key value pair inside the dict. And key value pairs syntactically close with ';')
                    pass
                # dict:
                if temp_tokens[-1][1] == '}':
                    # Proof that last key value pair in dict ends with ';'
                    # (= assert that second-to-last token is ';')
                    index = -2
                    # ..ok, line comments do not count .. skip them:
                    while re.match('^.*COMMENT.*$', str(temp_tokens[index][1])):
                        index -= 1
                    # ..but now: Does the last key value pair end with ';'?
                    if temp_tokens[index][1] not in ['{', ';', '}']:
                        # log error: Missing ';' after key value pair
                        logger.error(
                            'mis-spelled expression / missing \';\' around \"%s\"'
                            % ' '.join(t[1] for t in temp_tokens)
                        )

                # Parse the tokenized data struct, translate it into its type (list or dict),
                # and update parsed_list with the new list or dict.
                # Again, the code is type specific depending on whether the parsed data struct is a list or a dict.
                # list:
                if temp_tokens[0][1] == '(':
                    # Check whether the list is empty
                    if len(
                        temp_tokens
                    ) < 3:                                          # is empty (contains only the opening and the closing bracket)
                                                                    # add an empty list to parsed_list
                        parsed_list.append([])
                    else:                                           # has content
                                                                    # parse the nested list
                        nested_list = self._parse_tokenized_list(
                            dict, temp_tokens, level=level + 1
                        )                                           # (recursion)
                                                                    # add nested list to parsed_list
                        parsed_list.append(nested_list)
                                                                    #  dict:
                elif temp_tokens[0][1] == '{':
                                                                    # parse the nested dict
                    nested_dict = self._parse_tokenized_dict(
                        dict, temp_tokens[1:-1], level=level + 1
                    )
                                                                    # add nested dict to parsed_list
                    parsed_list.append(nested_dict)

                # All done: Identified data struct is parsed, translated into its corresponding type,
                # and local parsed_list is updated.
                # To close out and move on, fast-forward the index of tokens
                # to 'after' the data struct we just parsed:
                if last_index is not None:
                    token_index = last_index + 1

            # Single value type
            elif tokens[token_index][1] not in ['(', ')', ';']:
                value = self.parse_type(tokens[token_index][1])
                parsed_list.append(value)

            # -else = ';' or ')'
            else:
                pass

            # Iterate to next token
            token_index += 1

        # Return the parsed list
        return parsed_list

    def _find_companion(self, dict: CppDict, bracket: str) -> str:
        """Returns the companion bracket character for the passed in bracket character.

        Example: If you pass in '{', _find_companion() will return '}'  (and vice versa)
        """
        companion = ''
        for item in dict.brackets:
            if bracket == item[0]:
                companion = item[1]
            elif bracket == item[1]:
                companion = item[0]
        return companion

    def _insert_string_literals(self, dict: CppDict):
        """Substitutes STRINGLITERAL placeholders in the dict with the corresponding entry from dict.string_literals
        """

        for index, string_literal in dict.string_literals.items():
            # Properties of the expression to be evaluated
            placeholder = 'STRINGLITERAL%06i' % index   # STRINGLITERAL000000
                                                        # The entry from dict.string_literals is parsed once again, so that entries representing single value native types
                                                        # (such as bool ,None, int, float) are transformed to its native type, accordingly.
            value = self.parse_type(string_literal)

            # Replace all occurences of placeholder within the dictionary with the original string literal.
            # Note: As find_global_key() is non-greedy and returns the key of only the first occurance of placeholder it finds,
            # we need to loop until we found and replaced all occurances of placholder in the dict
            # and find_global_key() does not find any more occurances.
            while global_key := dict.find_global_key(query=placeholder):
                # Back insert the string literal
                dict.set_global_key(global_key, value)
        dict.string_literals.clear()
        return

    def _clean(self, dict: CppDict):
        """Removes CppFormatter / CppParser specific internal keys from dict.

        Removes keys written by CppFormatter for documentation purposes
        but which shall not be created as keys in dict.data.
        In specific, it is the following two keys that get deleted if existing:
        _variables
        _includes
        """
        if '_variables' in dict.data.keys():
            del dict.data['_variables']
        if '_includes' in dict.data.keys():
            del dict.data['_includes']


class FoamParser(CppParser):
    """Parser to deserialize a string in OpenFOAM dictionary format into a CppDict.
    """

    def __init__(self):
        '''
        Implementation specific default configuration of FoamParser
        '''
        # Invoke base class constructor
        super().__init__()

    def parse_string(
        self,
        string: str,
        target_dict: CppDict,
        comments: bool = True,
    ) -> CppDict:
        # sourcery skip: inline-immediately-returned-variable
        """Parses a string in OpenFOAM dictionary format and deserializes it into a CppDict.

        Parameters
        ----------
        string : str
            the string to be parsed (i.e. the content of the file that had been read using parse_file())
        target_dict : CppDict
            the target dict the parsed dict file shall be merged into
        comments : bool, optional
            reads comments, by default True

        Returns
        -------
        CppDict
            the parsed dict
        """

        # +++CALL BASE CLASS IMPLEMENTATION+++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        parsed_dict = super().parse_string(string, target_dict, comments)

        # +++MERGE PARSED DICTIONARY INTO TARGET DICTIONARY+++++++++++++++++++++++++++++++++++++++++
        # target_dict.merge(parsed_dict)  # Not necessary. Done in base class CppDictParser already.

        # +++RETURN PARSED DICTIONARY+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        return parsed_dict


class JsonParser(Parser):
    """Parser to deserialize a string in JSON dictionary format into a CppDict.
    """

    def __init__(self):
        '''
        Implementation specific default configuration of JsonParser
        '''
        # Invoke base class constructor
        super().__init__()

    def parse_string(
        self,
        string: str,
        target_dict: CppDict,
        comments: bool = True,
    ) -> CppDict:
        """Parses a string in JSON dictionary format and deserializes it into a CppDict.

        Parameters
        ----------
        string : str
            the string to be parsed (i.e. the content of the file that had been read using parse_file())
        target_dict : CppDict
            the target dict the parsed dict file shall be merged into
        comments : bool, optional
            reads comments, by default True

        Returns
        -------
        CppDict
            the parsed dict
        """
        import json

        # +++CALL BASE CLASS IMPLEMENTATION+++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        parsed_dict = super().parse_string(string, target_dict, comments)

        # +++PARSE DICTIONARY+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        parsed_dict.data = json.loads(string)
        self._extract_includes(parsed_dict)

        # +++MERGE PARSED DICTIONARY INTO TARGET DICTIONARY+++++++++++++++++++++++++++++++++++++++++
        target_dict.merge(parsed_dict)

        # +++RETURN PARSED DICTIONARY+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        return parsed_dict

    def _extract_includes(self, dict: CppDict):
        from copy import deepcopy
        keys = list(dict.data.keys())
        include_placeholder_keys = {}
        for key in keys:
            if isinstance(key, str) and re.search(r'^\s*#\s*include', key):
                include_file_name = str(dict[key])
                include_file_name = __class__.remove_quotes_from_string(include_file_name)

                include_file_path = Path.joinpath(dict.path, include_file_name)

                include_file_name_temp = include_file_name.replace('\\', '\\\\')
                include_directive = f"#include '{include_file_name_temp}'"

                ii = self.counter()
                dict.includes.update(
                    {ii: (include_directive, include_file_name, include_file_path)}
                )

                include_placeholder_keys[f'INCLUDE{ii:06d}'] = f'INCLUDE{ii:06d}'
                del dict[key]

        data_temp = deepcopy(dict.data)
        dict.data.clear()
        dict.data.update(include_placeholder_keys)
        dict.data.update(data_temp)

        return


class XmlParser(Parser):
    """Parser to deserialize a string in XML format into a CppDict.
    """

    def __init__(
        self,
        add_node_numbering: bool = True,
    ):
        '''
        Implementation specific default configuration of XmlParser
        '''
        # Invoke base class constructor
        super().__init__()
        # Save default configuration as attributes
        self.add_node_numbering = add_node_numbering

    def parse_string(
        self,
        string: str,
        target_dict: CppDict,
        comments: bool = True,
    ) -> CppDict:
        """Parses a string in XML format and deserializes it into a CppDict.

        Parameters
        ----------
        string : str
            the string to be parsed (i.e. the content of the file that had been read using parse_file())
        target_dict : CppDict
            the target dict the parsed dict file shall be merged into
        comments : bool, optional
            reads comments, by default True

        Returns
        -------
        CppDict
            the parsed dict
        """

        # +++CALL BASE CLASS IMPLEMENTATION+++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        parsed_dict = super().parse_string(string, target_dict, comments)

        # +++PARSE XML++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Default configuration
        namespaces = {'xs': 'https://www.w3.org/2009/XMLSchema/XMLSchema.xsd'}
        root_tag = 'NOTSPECIFIED'

        # Create XML parser
        parser = ETCompatXMLParser()
        # Read root element from XML string
        root_element = fromstring(string.encode('utf-8'), parser)
        # Read root tag from XML string
        # there is a problem with .tag :
        # fromstring does not completely push all attributes into .attrib
        # only version, not xmlns
        # xmlns remains as {XMLNSCONTENT}ROOTTAG
        # re.sub to fix that temporarily
        # solution needed
        root_tag = re.sub(r'\{.*\}', '', root_element.tag) or root_tag
        # Read namespaces from XML string
        namespaces = dict(root_element.nsmap) or namespaces
        # Reformat None keys in namespaces to key 'None' (as string)
        temp_keys_copy = list(namespaces)
        for key in temp_keys_copy:
            if key is None:
                try:
                    value = namespaces[key]
                    del namespaces[key]
                    namespaces['None'] = value
                except Exception:
                    logger.exception(
                        "XmlParser.parseString(): Reformatting None keys in namespaces to key 'None' failed"
                    )

        # +++PARSE DICTIONARY+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Transform XML root node into dict

        parsed_dict.data = dict(self._parse_nodes(root_element, namespaces))

        # Document XML options inside the dict
        try:
            parsed_dict.data['_xmlOpts'] = {
                '_nameSpaces': namespaces,
                '_rootTag': root_tag,
                '_rootAttributes': dict(root_element.attrib.items()),
                '_addNodeNumbering': self.add_node_numbering,
            }

        except Exception:
            logger.exception('XmlParser.parseString(): Cannot write _nameSpaces to _xmlOpts')

        # +++MERGE PARSED DICTIONARY INTO TARGET DICTIONARY+++++++++++++++++++++++++++++++++++++++++
        target_dict.merge(parsed_dict)

        # +++RETURN PARSED DICTIONARY+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        return parsed_dict

    def _parse_nodes(
        self,
        root_element: Element,
        namespaces: MutableMapping,
    ) -> dict:
        """Recursively parses all nodes and saves the nodes' content in a dict
        """
        # Default case: Make all node tags temporarily unique by indexing them using BorgCounter
        node_tags = [
            re.sub(r'^(\{.*\})', '', node.tag)
            for node in root_element.findall('*', dict(namespaces))
        ]
        indexed_node_tags = []
        for item in node_tags:
            index = self.counter()
            indexed_node_tags.append(('%06i_%s' % (index, item), item))
        # indexed_node_tags.sort()

        content_dict = {}

        # Parse all nodes
        for index, item in enumerate(indexed_node_tags):
            # Read the node
            node_tag = item[0]

            if not self.add_node_numbering:
                # Non-default case: add_node_numbering has been set to False by the caller
                # -> remove the index again
                node_tag = re.sub(r'^\d{6}_', '', node_tag)

            nodes = root_element.findall('*', dict(namespaces))

            # The recursive part.
            # If there is a nested list, step in and resolve,
            # otherwise append node text to dict.
            if list(nodes[index]):
                # node contains child nodes
                content_dict[node_tag] = self._parse_nodes(nodes[index], namespaces)

            elif nodes[index].text is None or re.search(r'^[\s\n\r]*$', nodes[index].text or ''):
                # Node has either no content or contains an empty string <NODE ATTRIB=STRING><\NODE>
                # However, in order to be able to attach attributes to a node,
                # we still need to create a dict for the node, even if the node has no content.
                content_dict[node_tag] = {}

            else:
                # Node has content.
                # Unfortunately, multiline text is not properly represented by the text attribute of the Node class:
                # The indentation is made part of each line, which is nonsense.
                # Hence a small workaround: We split into lines, strip each line to get rid of the indentation,
                # and finally rebuild the text by concatenating the stripped lines.
                text = nodes[index].text
                if text is not None:
                    original_lines = text.splitlines(keepends=True)
                    stripped_lines = [line.strip() for line in original_lines]
                    text = ('\n'.join(stripped_lines)).strip()
                else:
                    text = ''
                content_dict[node_tag] = {'_content': text}

            # If the node contains attributes: Save the attributes
            # and merge with the contents
            if len(nodes[index].attrib) > 0:
                # Avoid empty strings in attributes
                # Might be substtituted by any kind of substitution later if required.
                attributes_dict = {
                    '_attributes':
                    {k: str(v)
                     for k, v in nodes[index].attrib.items()
                     if str(v) != ''}
                }

                if content_dict[node_tag] is None:
                    content_dict[node_tag] = attributes_dict
                else:
                    content_dict[node_tag].update(attributes_dict)

        # before returning the new dict, doublecheck that all of its elements are correctly typed.
        self.parse_types(content_dict)

        return content_dict
